package OJS.Midi;

import java.io.IOException;
import javax.sound.midi.InvalidMidiDataException;
import javax.sound.midi.MidiDevice;
import javax.sound.midi.MidiSystem;
import javax.sound.midi.MidiUnavailableException;
import javax.sound.midi.ShortMessage;
import javax.sound.midi.MidiMessage;
import javax.sound.midi.Receiver;
import javax.sound.midi.Transmitter;
import OJS.Network.*;


/**
 * @author shengz
 *
 * This class is the high level module which connects the input and output modules. 
 * Other modules can access midi devices through this module.   
 */

public class MidiIoController implements Receiver
{
	MidiOut output = null;
	Transmitter input = null;
	Transmitter input2 = null;
	Network n = null;
	boolean useDefaultSynth = false;
	int LOCAL_CHANNEL = 1;
	int REMOTE_CHANNEL = 2;
	
	/**
	 * Two transmitters are used to transmit input messages to output device
	 * and network buffer, respectively.
	 */
	
	public MidiIoController()
	{
		listDevices();
		
		// MidiOutput to play messages received from input
		output = new MidiOut(useDefaultSynth);
		
		// MidiInput capable transmitting midi msgs
		input = new MidiIn();
		
		// Second MidiInput transmitter for transmitting msgs to network
		input2 = new MidiIn();
		
		// Route input MIDI messages to output 
		input.setReceiver(output);
		
		// Route input MIDI messages to myself 
		input2.setReceiver(this);

	}	
	
	/**
	 * Send a note on and a note off messages to the output device.
	 */
	public void noteTest()
	{
		ShortMessage onMessage = null;		
	    ShortMessage offMessage = null;
	    
		try 
		{		
			onMessage = new ShortMessage();
			offMessage = new ShortMessage();
			out("setting on message...");	
			onMessage.setMessage(ShortMessage.NOTE_ON, 0, 60, 120);
			out("set off message...");
			offMessage.setMessage(ShortMessage.NOTE_OFF, 0, 60, 0);
			out("set on message...");
			out("On Msg: " + onMessage.getStatus() + " " + onMessage.getData1() + " " + onMessage.getData2());
			out("Off Msg: " + offMessage.getStatus() + " " + offMessage.getData1() + " " + offMessage.getData2());
						
		}
		catch (InvalidMidiDataException e)
		{
		}
		
		out("sending on message...");
		output.send(onMessage, -1);
		try
		{
			Thread.sleep(100);
		}
		catch (InterruptedException e)
		{
		}
		
		out("sending off message...");
		output.send(offMessage, -1);
		output.close();
	}
	
	/**
	 * Enumerate all the midi devices and print to console.
	 */
	public static void listDevices()
	{
	
		out("Available MIDI Devices:");
	
		MidiDevice.Info[]	aInfos = MidiSystem.getMidiDeviceInfo();
		for (int i = 0; i < aInfos.length; i++)
		{
			try
			{
				MidiDevice	device = MidiSystem.getMidiDevice(aInfos[i]);
				boolean		bAllowsInput = (device.getMaxTransmitters() != 0);
				boolean		bAllowsOutput = (device.getMaxReceivers() != 0);

				out("" + i + "  "
					+ (bAllowsInput?"IN ":"   ")
					+ (bAllowsOutput?"OUT  ":"     ")
					+ aInfos[i].getName() + ", "
					+ aInfos[i].getVendor() + ", "
					+ aInfos[i].getVersion() + ", "
					+ aInfos[i].getDescription());
			}				
			catch (MidiUnavailableException e)
			{
			}
		}
		if (aInfos.length == 0)
		{
			out("[No devices available]");
		}
	}
	
	/**
	 * Send all the notes in a MidiSnapShot to output device with their delay values
	 * @param s - MidiSnapShot data, stores notes and their delays
	 */
	public void sendSnapShot(MidiSnapShot s)
	{
		int time = 0;		
		for(int i = 0; i < s.size; i++)
		{		
			output.SendRemote(s.messages[i], s.times[i]-time);
			time = s.times[i];			
		}
	}

	/**
	 * Send a message to output device. Messege's channel is over written to 
	 * the local channel value.
	 * @param - message to be sent
	 * @param t - delay in miliseconds before the message is executed
	 */
	public void play(ShortMessage message, long t)
	{
		if(message instanceof ShortMessage)
			// Ignore system messages 
			
			if( (((ShortMessage)message).getCommand() != 0xF0))
			{	
				try{
					ShortMessage s = new ShortMessage();
					s.setMessage(((ShortMessage)message).getCommand(), 
								LOCAL_CHANNEL, 
								((ShortMessage)message).getData1(),
								((ShortMessage)message).getData2());
				
					output.send(s, t);
				} catch (InvalidMidiDataException e){}
			}		
	}	
	
	/**
	 * @param n - network object to receiver messages
	 */
	public void setNetwork(Network n)
	{
		this.n = n;
	}
	
	public void close()
	{
		input.close();
		input2.close();
		output.close();
	}
	
	/**
	 * Send the midi message to the network buffer. Messege's channel is over written to 
	 * the local channel value.
	 * @see javax.sound.midi.Receiver#send(javax.sound.midi.MidiMessage, long)
	 */
	public void send(MidiMessage message, long t)
	{			
		
		if(message instanceof ShortMessage)
			// Ignore system messages 
			
			if( (((ShortMessage)message).getCommand() != 0xF0))
			{	
				try{
					ShortMessage s = new ShortMessage();
					s.setMessage(((ShortMessage)message).getCommand(), 
								LOCAL_CHANNEL, 
								((ShortMessage)message).getData1(),
								((ShortMessage)message).getData2());
				
					n.bufferMidi(s, t);
				} catch (InvalidMidiDataException e){}
			}
	}
		
	private static void out(String strMessage)
	{
		System.out.println(strMessage);
	}	
}
